//
// exporter.cpp
//
// Copyright 2004 by the authors.  All rights reserved.
//
// Authors: Tero Hasu <tero.hasu@hut.fi>
//

#include "exporter.h"
#include <msvuids.h>
#include <smut.h>
#include <txtrich.h> // etext.lib

// not included in SDK 2.0, get from Symbian website;
// some system headers must be edited to work around
// still missing header files --
// or can download SDK 2.1, which apparently has
// full SMS headers
#include <smuthdr.h>


enum TExporterPanicCodes
{
  EPanicReserved = 0,
  EAssumedConverterAvailable,
  EAssumedAllConvertible,
  EAssumedNoAttachments,
  EAssumedNotDeleted,
  EAssumedNoMarkup,
  EUnexpectedRecipientStatus,
  EUnexpectedRequestState
};

static void EnginePanic(TInt aReason)
{
  _LIT(KPanicCat, "smsexporter");
  User::Panic(KPanicCat, aReason);
}



CExporterEngine* CExporterEngine::NewL(RFs& aFs)
{
  CExporterEngine* o = new (ELeave) CExporterEngine(aFs);
  CleanupStack::PushL(o);
  o->ConstructL();
  CleanupStack::Pop();
  return o;
}

CExporterEngine::~CExporterEngine()
{
  Cancel();

  // don't know if need to check, but let us be safe
  if (iFile.SubSessionHandle() != 0)
    iFile.Close();

  delete iList;
  delete iMsvEntry;
  delete iSmsClientMtm;
  delete iClientMtmRegistry;
  delete iMsvSession;
  delete iCnvCharacterSetConverter;
  delete iContactDatabase;
  delete iContactBuffer;

#if FEATURE_LOGGING
  if (iLogger) {
    iLogger->CloseLog();
    iLogger->Close();
    delete iLogger;
  }
#endif
}

CExporterEngine::CExporterEngine(RFs& aFs) :
    CActive(EPriorityStandard), iFs(aFs)
{
  CActiveScheduler::Add(this);
  iComplete = ETrue;
}

void CExporterEngine::ConstructL()
{
#if FEATURE_LOGGING
  iLogger = new (ELeave) RFileLogger;
  if (iLogger->Connect() != KErrNone) {
    delete iLogger;
    iLogger = NULL;
  } else {
    _LIT(KLoggerDir, "smsexporter");
    _LIT(KLoggerFile, "smsexporter.txt");
    iLogger->CreateLog(KLoggerDir, KLoggerFile, EFileLoggingModeOverwrite);
    iLogger->Write(_L("logging started"));
  }
#endif

  iContactDatabase = CContactDatabase::OpenL();
  iContactBuffer = HBufC::NewL(50);

  iCnvCharacterSetConverter = CCnvCharacterSetConverter::NewL();
  CCnvCharacterSetConverter::TAvailability avail =
    iCnvCharacterSetConverter->PrepareToConvertToOrFromL
    (KCharacterSetIdentifierUtf8, iFs);
  if (avail == CCnvCharacterSetConverter::ENotAvailable)
    User::Leave(KErrGeneral);
  else if (avail != CCnvCharacterSetConverter::EAvailable)
    EnginePanic(EAssumedConverterAvailable);
  
  //iMsvSession = CMsvSession::OpenAsyncL(*this);
  iMsvSession = CMsvSession::OpenSyncL(*this);

  iClientMtmRegistry = CClientMtmRegistry::NewL(*iMsvSession);
  iSmsClientMtm = static_cast<CSmsClientMtm*>
    (iClientMtmRegistry->NewMtmL(KUidMsgTypeSMS));

  // will store our selection here
  iList = new (ELeave) CMsvEntrySelection();

  TMsvId inboxId = KMsvGlobalInBoxIndexEntryId;
  iMsvEntry = CMsvEntry::NewL(*iMsvSession, inboxId,
			      TMsvSelectionOrdering());

  CMsvEntrySelection* sel; // is-a CArrayFixFlat<TMsvId>
  TInt count;

  // inbox
  iMsvEntry->SetEntryL(inboxId);
  sel = iMsvEntry->ChildrenWithMtmL(KUidMsgTypeSMS);
  count = sel->Count();
  if (count > 0) {
    CleanupStack::PushL(sel);
    const TMsvId* back = sel->Back(0);
    iList->AppendL(back, count);
    CleanupStack::Pop(); // sel
  }
  delete sel;
  sel = NULL;
#if FEATURE_LOGGING
  if (iLogger) {
      _LIT8(KInboxNum, "%d Inbox items");
      iLogger->WriteFormat(KInboxNum, count);
  }
#endif

  // sent
  TMsvId sentId = KMsvSentEntryId;
  iMsvEntry->SetEntryL(sentId);
  sel = iMsvEntry->ChildrenWithMtmL(KUidMsgTypeSMS);
  count = sel->Count();
  if (count > 0) {
    CleanupStack::PushL(sel);
    const TMsvId* back = sel->Back(0);
    iList->AppendL(back, count);
    CleanupStack::Pop(); // sel
  }
  delete sel;
  sel = NULL;
#if FEATURE_LOGGING
  if (iLogger) {
      _LIT8(KSentNum, "%d Sent items");
      iLogger->WriteFormat(KSentNum, count);
  }
#endif

  /*
  if (iLogger)
    for (TInt i=0; i<iList->Count(); i++) {
      TMsvId& msvId = iList->At(i);
      _LIT8(KLogId, "ID %d");
      iLogger->WriteFormat(KLogId, msvId);
      }*/
}

void CExporterEngine::DeleteMessagesL(const TCallBack& aCallBack)
{
  iComplete = EFalse;
  iDeleting = ETrue;
  iNextIndex = 0;

  iCallBack = aCallBack;
  iStatus = KRequestPending;
  SetActive();
  TRequestStatus* status = &iStatus;
  User::RequestComplete(status, KErrNone);
}

void CExporterEngine::ExportMessagesL(const TCallBack& aCallBack)
{
  TInt drive;
#ifdef __WINS__
  drive = EDriveC;
#else
  TVolumeInfo dummy;
  drive = EDriveE;
  if (iFs.Volume(dummy, drive))
    drive = EDriveC;
#endif

  TChar driveChar;
  User::LeaveIfError(iFs.DriveToChar(drive, driveChar));

  TFileName fileName;
  TTime now;
  now.HomeTime();
  _LIT(KNowFmt, "%F%Y-%M-%D");
  TBuf<100> nowText;
  now.FormatL(nowText, KNowFmt);
  _LIT(KFnameFmt, "%c:\\sms-messages-%S.utf8.txt");
  fileName.Format(KFnameFmt, static_cast<TUint>(driveChar), &nowText);

  User::LeaveIfError(iFile.Replace(iFs, fileName, EFileWrite));

  iComplete = EFalse;
  iDeleting = EFalse;
  iNextIndex = 0;

  iCallBack = aCallBack;
  iStatus = KRequestPending;
  SetActive();
  TRequestStatus* status = &iStatus;
  User::RequestComplete(status, KErrNone);
}

void CExporterEngine::ConvertFromUnicode
(TDes8& aForeign, const TDesC16& aUnicode) const
{
  // now convert into a known encoding (say UTF-8)
  // and append to a file, along with any header
  // info we wish to store; probably want some cool
  // separator after the message, too
  TInt error = iCnvCharacterSetConverter->ConvertFromUnicode
    (aForeign, aUnicode);
  if (error)
    // either did not fit into buffer, which is not
    // as we assumed, or could not convert everything,
    // although we are assuming we can since we're using
    // a Unicode encoding
    EnginePanic(EAssumedAllConvertible);
}

void CExporterEngine::RunL()
{
  TBool hasAttachment = EFalse;

  if ((iStatus.Int() == KErrNone) && (iNextIndex < iList->Count())) {
    TMsvId& msvId = iList->At(iNextIndex);

    TMsvId service; // will not use
    TMsvEntry msvEntry; // summary info for a message
    User::LeaveIfError(iMsvSession->GetEntry(msvId, service, msvEntry));
    if (msvEntry.Attachment()) {
      // we were assuming short messages do not have attachments,
      // but it that they sometimes are marked as having so,
      // for instance when sending VCARDs;
      // but the content is still in body text, so having
      // an "attachment is of little significance";
      // will just make a note of it having been marked so

      //EnginePanic(EAssumedNoAttachments);

      hasAttachment = ETrue;
    }
    if (msvEntry.Deleted()) {
      // ...and that they do not stick around when deleted
      EnginePanic(EAssumedNotDeleted);
    }
    // there are many other flags, but we ignore them,
    // as they are of little interest when saving messages

    if (iDeleting) {
      // like a Unix filesystem, so must delete entry from parent node;
      // we need nothing SMS-specific here, so no need to use the SMS
      // client MTM
      TMsvId parentId = msvEntry.Parent();
      iMsvEntry->SetEntryL(parentId);
      iMsvEntry->DeleteL(msvId);
    } else {
      iSmsClientMtm->SwitchCurrentEntryL(msvId);

      //  alternatively could do
      //iMsvEntry->SetEntryL(msvId);
      // sets context
      //iSmsClientMtm->SetCurrentEntryL(iMsvEntry);

      // loads it into cache;
      // if we modified it and wanted to save, would use SaveMessageL;
      // after loading, we can query body text and other attributes
      iSmsClientMtm->LoadMessageL();




      CRichText& richText = iSmsClientMtm->Body();
      if (richText.HasMarkupData())
	EnginePanic(EAssumedNoMarkup);
      
      /*
      TCharFormat format;
      TPtrC bodyText;
      // note that we are using non-segmented storage,
      // so this should get everything, assuming everything
      // has is in the same style, and we asserted that
      // above, I think -- but in practice this does
      // not appear to get the whole message
      richText.GetChars(bodyText, format, 0);
      TInt bodyLen = bodyText.Length();
      */
      TInt bodyLen = richText.DocumentLength();





      // we could consider using a fixed-size buffer,
      // but since it is possible to send "virtual" SMSs
      // consisting of far more than 160 characters,
      // that would not always work; here trying to
      // guesstimate how much space we might need
      // at most
      TInt maxBuf = bodyLen * 4;
      if (maxBuf < 256)
	maxBuf = 256;
      HBufC8* hbuf8 = HBufC8::NewLC(maxBuf);
      TPtr8 buf8(hbuf8->Des());

      // Some message content uses CRLF, and some LF only. To fully
      // match that we would want to do translation, but then we would
      // essentially be losing information.
      _LIT8(KNewline, "\n");
      _LIT8(KSpace, " ");
      _LIT8(KLeft, "<");
      _LIT8(KRight, ">");

      _LIT(KTimeFormat, "%F%Y.%M.%D %H:%T:%S.%C");

      TBuf<100> formatBuf;






      _LIT(KIdFormat, "Message-ID: %d");
      formatBuf.Format(KIdFormat, msvId);
      ConvertFromUnicode(buf8, formatBuf);
      User::LeaveIfError(iFile.Write(buf8));
      User::LeaveIfError(iFile.Write(KNewline));

      /*
      _LIT8(KDescriptionLabel, "Description: ");
      User::LeaveIfError(iFile.Write(KDescriptionLabel));
      ConvertFromUnicode(buf8, msvEntry.iDescription);
      User::LeaveIfError(iFile.Write(buf8));
      User::LeaveIfError(iFile.Write(KNewline));

      _LIT8(KDetailsLabel, "Details: ");
      User::LeaveIfError(iFile.Write(KDetailsLabel));
      ConvertFromUnicode(buf8, msvEntry.iDetails);
      User::LeaveIfError(iFile.Write(buf8));
      User::LeaveIfError(iFile.Write(KNewline));
      */

      _LIT8(KDateLabel, "Date: ");
      User::LeaveIfError(iFile.Write(KDateLabel));
      msvEntry.iDate.FormatL(formatBuf, KTimeFormat);
      ConvertFromUnicode(buf8, formatBuf);
      User::LeaveIfError(iFile.Write(buf8));
      User::LeaveIfError(iFile.Write(KNewline));


      if (hasAttachment) {
	_LIT8(fmtLit, "Content-Disposition: attachment\n");
	User::LeaveIfError(iFile.Write(fmtLit));
      }













      const CSmsHeader& smsHeader = iSmsClientMtm->SmsHeader();
      const CArrayPtrFlat<CSmsNumber>& recipients = smsHeader.Recipients();

      // if there is a recipient, then the from address appears
      // to be the same as the recipient address, which does
      // not make sense, but whatever
      if (recipients.Count() == 0) {
	_LIT8(KFromLabel, "From: ");
	TPtrC fromAddress = smsHeader.FromAddress();

	TBool gotName = EFalse;
	TPtr contactBuffer(iContactBuffer->Des());
	HBufC* fromName = GetContactNameByPhoneNoL(*iContactDatabase,
						   gotName,
						   contactBuffer,
						   fromAddress);
	if (gotName) {
	  if (fromName) {
	    delete iContactBuffer;
	    iContactBuffer = fromName;
	  }
	}

	User::LeaveIfError(iFile.Write(KFromLabel));
	if (gotName) {
	  ConvertFromUnicode(buf8, *iContactBuffer);
	  User::LeaveIfError(iFile.Write(buf8));
	  User::LeaveIfError(iFile.Write(KSpace));
	  User::LeaveIfError(iFile.Write(KLeft));
	}
	ConvertFromUnicode(buf8, fromAddress);
	User::LeaveIfError(iFile.Write(buf8));
	if (gotName) {
	  User::LeaveIfError(iFile.Write(KRight));
	}
	User::LeaveIfError(iFile.Write(KNewline));
      }

      for (TInt i=0; i<recipients.Count(); i++) {
	const CSmsNumber* smsNumber = recipients.At(i);

	_LIT8(KToLabel, "To: ");
	User::LeaveIfError(iFile.Write(KToLabel));

	TPtrC toName = smsNumber->Name();
	ConvertFromUnicode(buf8, toName);
	User::LeaveIfError(iFile.Write(buf8));
	User::LeaveIfError(iFile.Write(KSpace));
	User::LeaveIfError(iFile.Write(KLeft));

	TPtrC toAddress = smsNumber->Address();
	ConvertFromUnicode(buf8, toAddress);
	User::LeaveIfError(iFile.Write(buf8));

	User::LeaveIfError(iFile.Write(KRight));
	User::LeaveIfError(iFile.Write(KSpace));

	CMsvRecipient::TRecipientStatus recipientStatus = smsNumber->Status();
	if (recipientStatus == CMsvRecipient::ENotYetSent) {
	  _LIT8(KNotYetSentText, "not yet sent");
	  User::LeaveIfError(iFile.Write(KNotYetSentText));
	} else if (recipientStatus == CMsvRecipient::ESentSuccessfully) {
	  _LIT8(KSentSuccessfullyText, "sent OK");
	  User::LeaveIfError(iFile.Write(KSentSuccessfullyText));
	} else if (recipientStatus == CMsvRecipient::EFailedToSend) {
	  _LIT8(KFailedToSendText, "sending failed");
	  User::LeaveIfError(iFile.Write(KFailedToSendText));
	} else {
	  EnginePanic(EUnexpectedRecipientStatus);
	}

	User::LeaveIfError(iFile.Write(KSpace));
	_LIT8(KAtTime, "at");
	User::LeaveIfError(iFile.Write(KAtTime));
	User::LeaveIfError(iFile.Write(KSpace));
	const TTime& recTime = smsNumber->Time();
	// using a locale independent, easy to parse
	// format -- this should make it easy to write (and share)
	// scripts that convert into some other format
	recTime.FormatL(formatBuf, KTimeFormat);
	ConvertFromUnicode(buf8, formatBuf);
	User::LeaveIfError(iFile.Write(buf8));

	User::LeaveIfError(iFile.Write(KNewline));
      }

      // empty line between header and body
      User::LeaveIfError(iFile.Write(KNewline));






      /*
      ConvertFromUnicode(buf8, bodyText);
      User::LeaveIfError(iFile.Write(buf8));
      */
      HBufC* hbuf16 = HBufC::NewLC(bodyLen);
      TPtr buf16(hbuf16->Des());
      richText.Extract(buf16);
      ConvertFromUnicode(buf8, buf16);
      CleanupStack::PopAndDestroy(); // hbuf16
      User::LeaveIfError(iFile.Write(buf8));



      User::LeaveIfError(iFile.Write(KNewline));
      User::LeaveIfError(iFile.Write(KNewline)); // empty line






      CleanupStack::PopAndDestroy(); // hbuf8

      /* not like this --
	 too low level, and uses protected methods

	 // a store for an individual message
	 CMsvStore* store = iMsvEntry->ReadStoreL();
	 CleanupStack::PushL(store);
	 if (store->HasBodyTextL()) {
	 //CRichText* richText = CRichText::
	 //NewL(iEikonEnv.SystemParaFormatLayerL(),
	 //     iEikonEnv.SystemCharFormatLayerL(),
	 //     CEditableText::EFlatStorage);
	 //CleanupStack::PushL(richText);
	 //store->RestoreBodyTextL(*richText);

	 //CleanupStack::PopAndDestroy(); // richText
	 } else {
	 // presumably an SMS should always contain something
	 User::Invariant();
	 }
	 CleanupStack::PopAndDestroy(); // store
      */

    }

    iStatus = KRequestPending;
    SetActive();
    TRequestStatus* status = &iStatus;
    User::RequestComplete(status, KErrNone);

    iNextIndex++;
  } else {
    if (!iDeleting)
      iFile.Close();

    // all done, whether failed or not
    iComplete = ETrue;
  }

  // caller can check iComplete to determine if this
  // is just a progress report, and iStatus to see
  // what happened if iComplete is true;
  // assuming this call does not leave
  iCallBack.CallBack();
}

TInt CExporterEngine::RunError(TInt aError)
{
  if (IsActive())
    // assuming did not make a new request yet
    EnginePanic(EUnexpectedRequestState);
  
  iStatus = KRequestPending;
  SetActive();
  TRequestStatus* status = &iStatus;
  User::RequestComplete(status, aError);

#if FEATURE_LOGGING
  if (iLogger) {
    _LIT(KErrorNumText, "error %d occurred");
    iLogger->WriteFormat(KErrorNumText, aError);
  }
#endif

  // to indicate that we handled the error,
  // and the active scheduler does not need to
  return KErrNone;
}

void CExporterEngine::DoCancel()
{
  if (iStatus == KRequestPending)
    {
      // should never need to do this as we always
      // complete as soon as we make active
      TRequestStatus* status = &iStatus;
      User::RequestComplete(status, KErrCancel);

#if FEATURE_LOGGING
      if (iLogger)
	iLogger->Write(_L("cancelled"));
#endif
    }
}

/** note that we are not interested in any folder content
    change events -- we just make a list of the stuff
    we want to export, initially, and then we stick with
    that list and assume all those items will stay
    available until all exported
*/
void CExporterEngine::HandleSessionEventL(TMsvSessionEvent aEvent,
					  TAny* /*aArg1*/,
					  TAny* /*aArg2*/,
					  TAny* /*aArg3*/)
{
  if (aEvent == EMsvServerReady)
    {
      /*
      iClientMtmRegistry = CClientMtmRegistry::NewL(*iMsvSession);
      iSmsClientMtm = static_cast<CSmsClientMtm*>
	(iClientMtmRegistry->NewMtmL(KUidMsgTypeSMS));
      */
    }
  else if (aEvent == EMsvServerFailedToStart)
    {
      //TInt error = *static_cast<TInt*>(aArg1);
      //EnginePanic(EPanicCouldNotConnect);
    }
}

